==> anygraph-0.1.0.tar.gz: anygraph-0.1.0/anygraph/linkers.py <==
from collections.abc import MutableSet
from operator import attrgetter

from anygraph.tools import unique_name
from anygraph.visitors import Iterator, Visitor


class DelegateSet(MutableSet):

    def __init__(self, owner, linker, get_id):
        super().__init__()
        self.targets = {}
        self.owner = owner
        self.linker = linker
        self.get_id = get_id

    def __len__(self):
        return len(self.targets)

    def __iter__(self):
        return iter(self.targets.values())

    def __contains__(self, target):
        return self.get_id(target) in self.targets

    def add(self, target):
        if target is not None and self.get_id(target) not in self.targets:
            self.linker._check(self.owner, target)
            self.linker._on_link(self.owner, target)
            self.linker._link(self.owner, target)

    def discard(self, target):
        if target is not None and self.get_id(target) in self.targets:
            self.linker._unlink(self.owner, target)
            self.linker._on_unlink(self.owner, target)

    def update(self, targets):
        for target in targets:
            self.add(target)

    def replace(self, targets):
        self.clear()
        self.update(targets)

    def _set(self, target):
        self.targets[self.get_id(target)] = target

    def _del(self, target):
        self.targets.pop(self.get_id(target), None)

    def __str__(self):
        return f"{{{', '.join(map(str, self))}}}"

    def __repr__(self):
        return f"{{{', '.join(map(repr, self))}}}"


class DelegateMap(DelegateSet):

    def __getitem__(self, key):
        return self.targets[key]

    def keys(self):
        return self.targets.keys()

    def values(self):
        return self.targets.values()

    def items(self):
        return self.targets.items()

    def get(self, key, _default=None):
        return self.targets.get(key, _default)


class DelegateNamedMap(DelegateMap):
    def add(self, target):
        self.set_name(target)
        super().add(target)

    def set_name(self, target):
        target.name = unique_name(
            space=self.targets.keys(),
            base=type(target).__name__.lower(),
            name=self.get_id(target)
        )


class BaseLinker(object):
    """
    Baseclass for One and Many descriptors, with shared functionality.
    """
    get_id = id  # default

    def __init__(self, reverse_name=None, cyclic=True, to_self=True, on_link=None, on_unlink=None, get_id=None):
        """
        :param reverse_name: optional name of the reverse realtionship
        :param cyclic: whether the graph is allowed to be cyclic
        :param to_self: whether an node the graph is allowed to connect ot itself
        :param on_link(obj, next_obj): optional callback called just before a connection is made
        :param on_unlink(obj, next_obj): optional callback called just before a connection is broken
        :param get_id(obj): optional alternative callback to uniquely identify nodes
        """
        self.reverse_name = reverse_name
        self.cyclic = cyclic
        self.to_self = to_self
        self._do_on_link = on_link
        self._do_on_unlink = on_unlink
        self._get_id = get_id or self.get_id
        self.name = None

    def __set_name__(self, cls, name):
        self.name = name  # sets the name of the attribute when the interpreter first encounters the descriptor in a class

    def __get__(self, obj, cls=None):
        if obj is None:
            return self  # used to call methods on the descriptor (like 'iterate' below)
        try:
            return obj.__dict__[self.name]
        except KeyError:
            return self._init(obj)  # initializes attribute on first access

    def iterate(self, start_obj, cyclic=False, breadth_first=False):
        """
        iterate through the graph
        :param start_obj: starting object from which the graph is followed
        :param cyclic: whether nodes in the graph will be repeated
        :param breadth_first: depth_first iteration if false (default) else breadth_first iteration
        :yield: nodes in the graph
        """
        yield from Iterator(self.name).iterate(start_obj, cyclic=cyclic, breadth_first=breadth_first)

    __call__ = iterate  # shortcut to iteration

    def visit(self, start_obj, on_visit, cyclic=False, breadth_first=False):
        """ apply 'on_visit(obj, next_obj)' on the graph, other arguments as in 'iterate' """
        visitor = Visitor(self.name)
        return visitor(start_obj, on_visit, cyclic=cyclic, breadth_first=breadth_first)

    def build(self, start_obj, key='__iter__'):
        """
        Build a graph using a key function to find next nodes.
        :param start_obj: entry point object for which the graph is built
        :param key: if str: use attribute with name 'key' of encountered objects to find next objects,
                    if callable: must return the (an iterable of) the next objects in the graph
        :return: self: to chain a call to another method if you like.
        """
        build_on_visit = self._build_on_visit(key, _reg={})
        self.visit(start_obj, on_visit=build_on_visit, breadth_first=True)
        return self

    def gather(self, start_obj):  # bit slow
        """
        Gather all nodes in a graph, going forward and backward if reverse_name is defined.
        """
        get_id = self._get_id
        gathered = {get_id(obj): obj for obj in self.iterate(start_obj)}  # dict so no doubles but with non-hashable objs
        if self.reverse_name and self.name != self.reverse_name:
            for obj in gathered.values():
                for rev_obj in Iterator(self.reverse_name)(obj):  # follow the reverse relationship
                    gathered[get_id(rev_obj)] = rev_obj
        return list(gathered.values())

    def find(self, start_obj, filter):
        """ return objects that pass the filer callback """
        return [obj for obj in self.iterate(start_obj, breadth_first=True) if filter(obj)]

    def reachable(self, start_obj, target_obj):
        """ return whether target_obj can be reached from start_obj through the graph """
        iterator = Iterator(self.name)
        for obj in iterator(start_obj):
            if target_obj in iterator.iter_object(obj):
                return True
        return False

    def walk(self, start_obj, key, on_visit=None):
        """
        iterate through the graph, with a key function selecting the next connected node
        :param start_obj: starting point in the graph
        :param key: function to pick the next node in the graph, as in 'next_node = key(node)'
        :param on_visit: optional callback applied to the visited node
        :return: yield nodes encountered in the graph one-by-one
        """
        yield from Iterator(self.name).walk(start_obj, key=key, on_visit=on_visit)

    def endpoints(self, start_obj):
        """ return a list of endpoint nodes (with no next node in the graph), starting from start_obj"""
        endpoints = []
        iterator = Iterator(self.name)
        for obj in iterator(start_obj):
            if not list(iterator.iter_object(obj)):
                endpoints.append(obj)
        return endpoints

    def is_cyclic(self, start_obj):
        """ returns whether any cycles are in the graph reachable from start_obj"""
        seen = set()
        cyclic = False

        def visit(obj):
            nonlocal cyclic
            obj_id = self._get_id(obj)
            if obj_id in seen:
                cyclic = True
                raise StopIteration
            seen.add(obj_id)

        self.visit(start_obj, on_visit=visit, cyclic=True)
        return cyclic

    def in_cycle(self, start_obj):
        """ return whether start_obj is in a cycle (whether it can be reached from itself)"""
        return self.reachable(start_obj, start_obj)

    def shortest_path(self, start_obj, target_obj, get_cost=None, heuristic=None):
        """
         Finds the shortest path through the graph from start_obj to target_obj
        :param start_obj: node to start from
        :param target_obj: node to which the path must be calculated
        :param get_cost(node, next_node): cost function: must return cost for following edge between node and next_node. Default
            results in a shortest path defined by the number of edges between start and end.
        :param heuristic(node, target_node): optional heuristic function to calculate an under estimate of the remaining
            cost from a node to the target node (often resulting in faster path_finding, using A*).
        :return: list of nodes of the shortest path
        """
        return Iterator(self.name).shortest_path(start_obj, target_obj,
                                                 get_cost=get_cost,
                                                 heuristic=heuristic)

    def _reverse(self, target):
        if self.reverse_name is None:
            return None
        return getattr(target.__class__, self.reverse_name)

    def _existing(self, obj, target):
        raise NotImplementedError

    def _creates_cycle(self, obj, target):
        if obj is target:
            return True
        if self.name == self.reverse_name:
            return True
        if target is None:
            return False
        if self._existing(obj, target):
            return False
        return self.reachable(target, obj)

    def _check(self, obj, target):
        if target is None:
            return
        reverse = self._reverse(target)
        if not (self.to_self and (not reverse or reverse.to_self)):  # reverse might be a different relationship
            if obj is target:
                raise ValueError(f"pointing '{self.name}' in {obj.__class__.__name__} back to self: 'to_self' is set to False")
        if not (self.cyclic and (not reverse or reverse.cyclic)):
            if self._creates_cycle(obj, target):
                raise ValueError(f"setting '{self.name}' in {obj.__class__.__name__} creates cycle: 'cyclic' is set to False")

    def _link(self, obj, target):
        self._set(obj, target)
        if self.reverse_name:
            self._reverse(target)._set(target, obj)

    def _unlink(self, obj, target=None):
        if target is not None:
            if self.reverse_name:
                self._reverse(target)._del(target, obj)
            self._del(obj, target)

    def _on_link(self, obj, target, _remote=False):
        if self._do_on_link:
            return self._do_on_link(obj, target)
        if self.reverse_name and not _remote:
            return self._reverse(target)._on_link(target, obj, True)

    def _on_unlink(self, obj, target, _remote=False):
        if self._do_on_unlink:
            return self._do_on_unlink(obj, target)
        if self.reverse_name and not _remote:
            return self._reverse(target)._on_unlink(target, obj, True)

    def _init(self, obj):
        raise NotImplementedError

    def _set(self, obj, value):
        raise NotImplementedError

    def _del(self, obj, target):
        raise NotImplementedError

    def _build_on_visit(self, key, _reg):
        raise NotImplementedError


class One(BaseLinker):

    def __set__(self, obj, target):
        self._check(obj, target)
        self._unlink(obj)
        if target is not None:
            if self.reverse_name:
                self._reverse(target)._unlink(target)
            self._on_link(obj, target)
            self._link(obj, target)

    def __delete__(self, obj):
        target = self.__get__(obj)
        self._unlink(obj, target)
        self._on_unlink(obj, target)

    def _existing(self, obj, target):
        return target is not None and self.__get__(obj) is target

    def _build_on_visit(self, key, _reg):
        get_id = self._get_id

        def visit(obj):
            if isinstance(key, str):
                value = getattr(obj, key)
            else:
                value = key(obj)
            id_ = get_id(value)
            if id_ in _reg:
                value = _reg[id_]
            else:
                _reg[id_] = value
            setattr(obj, self.name, value)
        return visit

    def _init(self, obj):
        obj.__dict__[self.name] = None

    def _set(self, obj, target):
        obj.__dict__[self.name] = target

    def _del(self, obj, target):
        obj.__dict__[self.name] = None

    def _unlink(self, obj, target=None):
        if target is None:
            target = self.__get__(obj)
        super()._unlink(obj, target)


class BaseMany(BaseLinker):
    many_class = None

    def __set__(self, obj, targets):
        self.__get__(obj).replace(targets)

    def __delete__(self, obj):
        self.__get__(obj).clear()

    def _existing(self, obj, target):
        return target is not None and target in self.__get__(obj)

    def _build_on_visit(self, key, _reg):
        get_id = self._get_id

        def visit(obj, key=key):
            if isinstance(key, str):
                key = getattr(obj.__class__, key)
            for value in key(obj):
                ident = get_id(value)
                if ident in _reg:
                    value = _reg[ident]
                else:
                    _reg[ident] = value
                getattr(obj, self.name).add(value)
        return visit

    def _init(self, obj):
        obj.__dict__[self.name] = self.many_class(obj, linker=self, get_id=self._get_id)
        return obj.__dict__[self.name]

    def _set(self, obj, target):
        self.__get__(obj)._set(target)

    def _del(self, obj, target):
        self.__get__(obj)._del(target)


class Many(BaseMany):
    many_class = DelegateSet


class ManyMap(BaseMany):
    many_class = DelegateMap


class ManyNamed(BaseMany):
    get_id = attrgetter('name')
    many_class = DelegateNamedMap


if __name__ == '__main__':
    pass

==> anygraph-0.1.1.tar.gz: anygraph-0.1.1/anygraph/linkers.py <==
from collections import deque
from collections.abc import MutableSet
from operator import attrgetter

from anygraph.tools import unique_name
from anygraph.visitors import Iterator, Visitor


class DelegateSet(MutableSet):

    def __init__(self, owner, linker, get_id):
        super().__init__()
        self.targets = {}
        self.owner = owner
        self.linker = linker
        self.get_id = get_id

    def __len__(self):
        return len(self.targets)

    def __iter__(self):
        return iter(self.targets.values())

    def __contains__(self, target):
        return self.get_id(target) in self.targets

    def add(self, target):
        if target is not None and self.get_id(target) not in self.targets:
            self.linker._check(self.owner, target)
            self.linker._on_link(self.owner, target)
            self.linker._link(self.owner, target)

    def discard(self, target):
        if target is not None and self.get_id(target) in self.targets:
            self.linker._unlink(self.owner, target)
            self.linker._on_unlink(self.owner, target)

    def update(self, targets):
        for target in targets:
            self.add(target)

    def replace(self, targets):
        self.clear()
        self.update(targets)

    def _set(self, target):
        self.targets[self.get_id(target)] = target

    def _del(self, target):
        self.targets.pop(self.get_id(target), None)

    def __str__(self):
        return f"{{{', '.join(map(str, self))}}}"

    def __repr__(self):
        return f"{{{', '.join(map(repr, self))}}}"


class DelegateMap(DelegateSet):

    def __getitem__(self, key):
        return self.targets[key]

    def keys(self):
        return self.targets.keys()

    def values(self):
        return self.targets.values()

    def items(self):
        return self.targets.items()

    def get(self, key, _default=None):
        return self.targets.get(key, _default)


class DelegateNamedMap(DelegateMap):
    def add(self, target):
        self.set_name(target)
        super().add(target)

    def set_name(self, target):
        target.name = unique_name(
            space=self.targets.keys(),
            base=type(target).__name__.lower(),
            name=self.get_id(target)
        )


class BaseLinker(object):
    """
    Baseclass for One and Many descriptors, with shared functionality.
    """
    get_id = id  # default

    def __init__(self, reverse_name=None, cyclic=True, to_self=True, on_link=None, on_unlink=None, get_id=None):
        """
        :param reverse_name: optional name of the reverse realtionship
        :param cyclic: whether the graph is allowed to be cyclic
        :param to_self: whether an node the graph is allowed to connect ot itself
        :param on_link(obj, next_obj): optional callback called just before a connection is made
        :param on_unlink(obj, next_obj): optional callback called just before a connection is broken
        :param get_id(obj): optional alternative callback to uniquely identify nodes
        """
        self.reverse_name = reverse_name
        self.cyclic = cyclic
        self.to_self = to_self
        self._do_on_link = on_link
        self._do_on_unlink = on_unlink
        self._get_id = get_id or self.get_id
        self.name = None

    def __set_name__(self, cls, name):
        self.name = name  # sets the name of the attribute when the interpreter first encounters the descriptor in a class

    def __get__(self, obj, cls=None):
        if obj is None:
            return self  # used to call methods on the descriptor (like 'iterate' below)
        try:
            return obj.__dict__[self.name]
        except KeyError:
            return self._init(obj)  # initializes attribute on first access

    def iterate(self, start_obj, cyclic=False, breadth_first=False):
        """
        iterate through the graph
        :param start_obj: starting object from which the graph is followed
        :param cyclic: whether nodes in the graph will be repeated
        :param breadth_first: depth_first iteration if false (default) else breadth_first iteration
        :yield: nodes in the graph
        """
        yield from Iterator(self.name).iterate(start_obj, cyclic=cyclic, breadth_first=breadth_first)

    __call__ = iterate  # shortcut to iteration

    def visit(self, start_obj, on_visit, cyclic=False, breadth_first=False):
        """ apply 'on_visit(obj, next_obj)' on the graph, other arguments as in 'iterate' """
        visitor = Visitor(self.name)
        return visitor(start_obj, on_visit, cyclic=cyclic, breadth_first=breadth_first)

    def build(self, start_obj, key='__iter__'):
        """
        Build a graph using a key function to find next nodes.
        :param start_obj: entry point object for which the graph is built
        :param key: if str: use attribute with name 'key' of encountered objects to find next objects,
                    if callable: must return the (an iterable of) the next objects in the graph
        :return: self: to chain a call to another method if you like.
        """
        build_on_visit = self._build_on_visit(key, _reg={})
        self.visit(start_obj, on_visit=build_on_visit, breadth_first=True)
        return self

    def gather(self, start_obj):  # bit slow
        """
        Gather all nodes in a graph, going forward and backward if reverse_name is defined.
        """
        get_id = self._get_id

        forw_iterator = Iterator(self.name)
        back_iterator = Iterator(self.reverse_name)

        gathered = {}
        queue = deque([start_obj])

        while len(queue):
            obj = queue.popleft()
            gathered[get_id(obj)] = obj

            for forw_obj in forw_iterator.iter_object(obj):
                if get_id(forw_obj) not in gathered:
                    queue.append(forw_obj)

            for back_obj in back_iterator.iter_object(obj):
                if get_id(back_obj) not in gathered:
                    queue.append(back_obj)

        return list(gathered.values())

    def find(self, start_obj, filter):
        """ return objects that pass the filer callback """
        return [obj for obj in self.iterate(start_obj, breadth_first=True) if filter(obj)]

    def reachable(self, start_obj, target_obj):
        """ return whether target_obj can be reached from start_obj through the graph """
        iterator = Iterator(self.name)
        for obj in iterator(start_obj):
            if target_obj in iterator.iter_object(obj):
                return True
        return False

    def walk(self, start_obj, key, on_visit=None):
        """
        iterate through the graph, with a key function selecting the next connected node
        :param start_obj: starting point in the graph
        :param key: function to pick the next node in the graph, as in 'next_node = key(node)'
        :param on_visit: optional callback applied to the visited node
        :return: yield nodes encountered in the graph one-by-one
        """
        yield from Iterator(self.name).walk(start_obj, key=key, on_visit=on_visit)

    def endpoints(self, start_obj):
        """ return a list of endpoint nodes (with no next node in the graph), starting from start_obj"""
        endpoints = []
        iterator = Iterator(self.name)
        for obj in iterator(start_obj):
            if not list(iterator.iter_object(obj)):
                endpoints.append(obj)
        return endpoints

    def is_cyclic(self, start_obj):
        """ returns whether any cycles are in the graph reachable from start_obj"""
        seen = set()
        cyclic = False

        def visit(obj):
            nonlocal cyclic
            obj_id = self._get_id(obj)
            if obj_id in seen:
                cyclic = True
                raise StopIteration
            seen.add(obj_id)

        self.visit(start_obj, on_visit=visit, cyclic=True)
        return cyclic

    def in_cycle(self, start_obj):
        """ return whether start_obj is in a cycle (whether it can be reached from itself)"""
        return self.reachable(start_obj, start_obj)

    def shortest_path(self, start_obj, target_obj, get_cost=None, heuristic=None):
        """
         Finds the shortest path through the graph from start_obj to target_obj
        :param start_obj: node to start from
        :param target_obj: node to which the path must be calculated
        :param get_cost(node, next_node): cost function: must return cost for following edge between node and next_node. Default
            results in a shortest path defined by the number of edges between start and end.
        :param heuristic(node, target_node): optional heuristic function to calculate an under estimate of the remaining
            cost from a node to the target node (often resulting in faster path_finding, using A*).
        :return: list of nodes of the shortest path
        """
        return Iterator(self.name).shortest_path(start_obj, target_obj,
                                                 get_cost=get_cost,
                                                 heuristic=heuristic)

    def _reverse(self, target):
        if self.reverse_name is None:
            return None
        return getattr(target.__class__, self.reverse_name)

    def _existing(self, obj, target):
        raise NotImplementedError

    def _creates_cycle(self, obj, target):
        if obj is target:
            return True
        if self.name == self.reverse_name:
            return True
        if target is None:
            return False
        if self._existing(obj, target):
            return False
        return self.reachable(target, obj)

    def _check(self, obj, target):
        if target is None:
            return
        reverse = self._reverse(target)
        if not (self.to_self and (not reverse or reverse.to_self)):  # reverse might be a different relationship
            if obj is target:
                raise ValueError(f"pointing '{self.name}' in {obj.__class__.__name__} back to self: 'to_self' is set to False")
        if not (self.cyclic and (not reverse or reverse.cyclic)):
            if self._creates_cycle(obj, target):
                raise ValueError(f"setting '{self.name}' in {obj.__class__.__name__} creates cycle: 'cyclic' is set to False")

    def _link(self, obj, target):
        self._set(obj, target)
        if self.reverse_name:
            self._reverse(target)._set(target, obj)

    def _unlink(self, obj, target=None):
        if target is not None:
            if self.reverse_name:
                self._reverse(target)._del(target, obj)
            self._del(obj, target)

    def _on_link(self, obj, target, _remote=False):
        if self._do_on_link:
            return self._do_on_link(obj, target)
        if self.reverse_name and not _remote:
            return self._reverse(target)._on_link(target, obj, True)

    def _on_unlink(self, obj, target, _remote=False):
        if self._do_on_unlink:
            return self._do_on_unlink(obj, target)
        if self.reverse_name and not _remote:
            return self._reverse(target)._on_unlink(target, obj, True)

    def _init(self, obj):
        raise NotImplementedError

    def _set(self, obj, value):
        raise NotImplementedError

    def _del(self, obj, target):
        raise NotImplementedError

    def _build_on_visit(self, key, _reg):
        raise NotImplementedError


class One(BaseLinker):

    def __set__(self, obj, target):
        self._check(obj, target)
        self._unlink(obj)
        if target is not None:
            if self.reverse_name:
                self._reverse(target)._unlink(target)
            self._on_link(obj, target)
            self._link(obj, target)

    def __delete__(self, obj):
        target = self.__get__(obj)
        self._unlink(obj, target)
        self._on_unlink(obj, target)

    def _existing(self, obj, target):
        return target is not None and self.__get__(obj) is target

    def _build_on_visit(self, key, _reg):
        get_id = self._get_id

        def visit(obj):
            if isinstance(key, str):
                value = getattr(obj, key)
            else:
                value = key(obj)
            id_ = get_id(value)
            if id_ in _reg:
                value = _reg[id_]
            else:
                _reg[id_] = value
            setattr(obj, self.name, value)
        return visit

    def _init(self, obj):
        obj.__dict__[self.name] = None

    def _set(self, obj, target):
        obj.__dict__[self.name] = target

    def _del(self, obj, target):
        obj.__dict__[self.name] = None

    def _unlink(self, obj, target=None):
        if target is None:
            target = self.__get__(obj)
        super()._unlink(obj, target)


class BaseMany(BaseLinker):
    many_class = None

    def __set__(self, obj, targets):
        self.__get__(obj).replace(targets)

    def __delete__(self, obj):
        self.__get__(obj).clear()

    def _existing(self, obj, target):
        return target is not None and target in self.__get__(obj)

    def _build_on_visit(self, key, _reg):
        get_id = self._get_id

        def visit(obj, key=key):
            if isinstance(key, str):
                key = getattr(obj.__class__, key)
            for value in key(obj):
                ident = get_id(value)
                if ident in _reg:
                    value = _reg[ident]
                else:
                    _reg[ident] = value
                getattr(obj, self.name).add(value)
        return visit

    def _init(self, obj):
        obj.__dict__[self.name] = self.many_class(obj, linker=self, get_id=self._get_id)
        return obj.__dict__[self.name]

    def _set(self, obj, target):
        self.__get__(obj)._set(target)

    def _del(self, obj, target):
        self.__get__(obj)._del(target)


class Many(BaseMany):
    many_class = DelegateSet


class ManyMap(BaseMany):
    many_class = DelegateMap


class ManyNamed(BaseMany):
    get_id = attrgetter('name')
    many_class = DelegateNamedMap


if __name__ == '__main__':
    pass
